import json
import os.path
import platform
import subprocess

import requests

def isgoogle():
    # Check if Google
    r = requests.get('http://169.254.169.254/computeMetadata/v1/instance/id', headers={'Metadata-Flavor': 'Google'})
    if r.ok:
        vmid = r.text
        r = requests.get('http://169.254.169.254/computeMetadata/v1/instance/zone', headers={'Metadata-Flavor': 'Google'})
        if r.ok:
            zone = r.text.rpartition('/')[-1]
        return {'cloud': 'gcp', 'vmid': vmid, 'region': zone}
    return False

def isaws():
    r = requests.get('http://169.254.169.254/latest/dynamic/instance-identity/document', headers={'Metadata': 'true'})
    if r.ok:
        j = r.json()
        region = j['availabilityZone']
        vmid = j['instanceId']
        return {'cloud': 'aws', 'vmid': vmid, 'region': region}
    return False

def isazure():
    r = requests.get('http://169.254.169.254/metadata/instance?api-version=2021-02-01', headers={'Metadata': 'True'})
    if r.ok:
        j = r.json()
        region = j['compute']['location']
        vmid = j['compute']['vmId']
        # region = j['location']
        # vmid = j['vmId']
        return {'cloud': 'azure', 'vmid': vmid, 'region': region}
    return False

# res = {'cloud': 'cloud', 'vmid': 'vmid', 'region': 'region'}

# if os.path.exists('config.json'):
#     with open('config.json') as f:
#         res = json.load(f)
# else:


def get_cloud_info():
    res = None
    for func in [isaws, isazure, isgoogle]:
        try:
            res = func()
            if res:
                break
        except requests.exceptions.ConnectionError:
            continue
    if res:
        hostname = f'{res["cloud"]}_{res["region"]}_{res["vmid"]}'
        print(f'Found cloud metadata: {hostname}')
    else:
        hostname = platform.node()
        print(f'Unable to find any cloud metadata. Using hostname: {hostname}')
    return hostname




import os
import random
import subprocess
import tempfile
import time
import urllib.request
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import platform
from datetime import date

from file2 import fopen

from cloudtrace.trace.cloudinfo import get_cloud_info
from cloudtrace.trace.shuffle import shuf
from cloudtrace.trace.fasttrace import remote_notify
from cloudtrace.version import __version__

def new_filename(default_output, pps, ext, gzip=False, bzip2=False, hostname=None):
    if hostname is None:
        hostname = get_cloud_info()
    dirname, basename = os.path.split(default_output)
    if dirname:
        os.makedirs(dirname, exist_ok=True)
    if basename:
        basename += '.'
    timestamp = int(time.time())
    dt = date.fromtimestamp(timestamp)
    datestr = dt.strftime('%Y%m%d')
    filename = os.path.join(dirname, f'{basename}{hostname}.{datestr}.{timestamp}.{pps}.{ext}')
    if gzip:
        filename += '.gz'
    elif bzip2:
        filename += '.bz2'
    return filename

def print_duration(start, end):
    secs = end - start
    mins = secs / 60
    hours = mins / 60
    print(f'Duration: {hours:,.2f} h {mins:,.2f} m {secs:,.2f} s')

def main():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-f', '--input')
    group.add_argument('-i', '--addr', nargs='*')
    parser.add_argument('-p', '--pps', default=5000, type=int, help='Packets per second.')
    parser.add_argument('-o', '--default-output', required=True)
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-z', '--gzip', action='store_true')
    group.add_argument('-b', '--bzip2', action='store_true')
    parser.add_argument('--remote')
    parser.add_argument('--cycles', type=int, default=1)
    parser.add_argument('--random', action='store_true')
    parser.add_argument('--shuffle', action='store_true')
    parser.add_argument('-O', '--extension', choices=['warts', 'json'], default='json')
    subparsers = parser.add_subparsers()
    tparser = subparsers.add_parser('trace')
    tparser.set_defaults(cmd='trace')
    pparser = subparsers.add_parser('ping')
    pparser.set_defaults(cmd='ping')
    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=__version__))
    args, remaining = parser.parse_known_args()

    hostname = get_cloud_info()
    addr = urllib.request.urlopen('https://api.ipify.org/').read().decode('utf8')

    sccmd = args.cmd + ' ' + ' '.join(remaining)

    cycle = 0
    while args.cycles == 0 or cycle < args.cycles:
        infile = args.input
        if args.shuffle:
            shuf(infile, inplace=True)
        f = tempfile.NamedTemporaryFile(mode='wt', delete=False)
        tmp = f.name
        try:
            if args.input:
                with fopen(infile, 'rt') as g:
                    for line in g:
                        if args.random:
                            addr, _, _ = line.rpartition('.')
                            addr = f'{addr}.{random.randint(0, 255)}'
                        else:
                            addr = line.strip()
                        f.write(f'{addr}\n')
            else:
                f.writelines(f'{addr}\n' for addr in args.addr)
            f.close()

            ftype = args.extension
            filename = new_filename(args.default_output, args.pps, ftype, gzip=args.gzip, bzip2=args.bzip2, hostname=hostname)
            if args.gzip:
                write = f'| gzip > {filename}'
            elif args.bzip2:
                write = f'| bzip2 > {filename}'
            else:
                write = f'-o {filename}'
            cmd = f'sudo scamper -O {ftype} -p {args.pps} -l {addr} -M {hostname} -c "{sccmd}" -f {tmp} {write}'
            print(cmd)
            start = time.time()
            subprocess.run(cmd, shell=True, check=False)
            end = time.time()
            print_duration(start, end)
            if args.remote:
                dirname, basename = os.path.split(args.default_output)
                pattern = os.path.join(dirname, f'{basename}.warts*')
                remote_notify(pattern, args.remote)
            try:
                cycle += 1
            except OverflowError:
                cycle = 1
        finally:
            os.unlink(tmp)



import platform
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from multiprocessing import Pool

from pb_amarder.bar import Progress

from cloudtrace.read.convert import ConvertTrace

verbose = 0
fix = True

def pattern(infile, gzip=False, bzip2=False):
    if infile.endswith('.bz2'):
        outfile = infile[:-4]
    elif infile.endswith('.gz'):
        outfile = infile[:-3]
    else:
        outfile = infile
    outfile = '{}.jsonl'.format(outfile)
    if gzip:
        outfile += '.gz'
    elif bzip2:
        outfile += '.bz2'
    return outfile

def convert(infile, outfile):
    sidefile = (infile.rpartition('.')[0] if infile.endswith('.gz') or infile.endswith('.bz2') else infile) + '.sidecar'
    c = ConvertTrace(infile, outfile, sidefile)
    c.convert_trace(fix=fix, verbose=verbose)
    if verbose:
        print(c)

def convert_wrapper(args):
    infile, outfile = args
    convert(infile, outfile)

def main():
    global fix, verbose
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-r', '--raw', action='store_true')
    parser.add_argument('-v', '--verbose', action='count', default=0)
    parser.add_argument('-o', '--outfiles', nargs='+')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-z', '--gzip', action='store_true')
    group.add_argument('-b', '--bzip2', action='store_true')
    parser.add_argument('-p', '--processes', default=1, type=int)
    parser.add_argument('-i', '--infiles', nargs='+')
    parser.add_argument('-I', '--file')
    parser.add_argument('-H', '--hostname', default=platform.node())
    args = parser.parse_args()

    fix = not args.raw
    print('Fixing pcap file: {}'.format(fix))
    verbose = args.verbose

    infiles = args.infiles if args.infiles is not None else []
    if args.file:
        with open(args.file, 'rt') as f:
            for filename in f:
                infiles.append(filename.strip())

    if args.outfiles:
        if len(args.outfiles) != len(infiles):
            print('Outfiles must match number of infiles.')
            exit(1)
        outfiles = args.outfiles
    else:
        outfiles = []
        for infile in infiles:
            outfile = pattern(infile, gzip=args.gzip, bzip2=args.bzip2)
            outfiles.append(outfile)

    # print(outfiles)

    processes = min(args.processes, len(infiles))

    pb = Progress(len(infiles), message=f'Creating traceroutes with {processes} process')
    if processes == 1:
        for infile, outfile in pb.iterator(zip(infiles, outfiles)):
            convert(infile, outfile)
    else:
        with Pool(processes) as pool:
            for _ in pb.iterator(pool.imap_unordered(convert_wrapper, zip(infiles, outfiles))):
                pass

    # pb = Progress(len(infiles), message='Creating traceroutes')
    # if args.processes == 1:
    #     for t in pb.iterator(zip(args.infiles, outfiles)):
    #         convert(t)
    # else:
    #     with Pool(max(args.processes, len(infiles))) as pool:
    #         for _ in pb.iterator(pool.imap_unordered(convert, zip(infiles, outfiles))):
    #             pass


if __name__ == '__main__':
    main()



import json
import os.path
import sys
import traceback
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from collections import defaultdict, namedtuple, OrderedDict
from datetime import timedelta
from multiprocessing import Pool
from subprocess import Popen, PIPE
from sys import stderr
from typing import List, Dict
from socket import IPPROTO_ICMP, IPPROTO_UDP

from file2 import fopen
from pb_amarder import Progress

# {"type": "trace", "version": "0.1", "userid": 0, "method": "udp-paris", "src": "172.31.25.243", "dst": "99.198.241.1", "sport": 6179, "dport": 30000, "stop_reason": "GAPLIMIT", "stop_data": 0, "start": {"sec": 1666208958, "usec": 841377, "ftime": "fake"}, "hop_count": 5, "attempts": 1, "hoplimit": 32, "firsthop": 1, "wait": 15, "wait_probe": 0, "tos": 0, "probe_size": 64}

_fix = True
verbose = False
_timeout = 10

TraceInfo = namedtuple('TraceInfo', ['src', 'dst', 'proto'])

def remove_old(ts, times, reqs, timeout):
    sec = ts['ts_sec']
    todel = []
    for tinfo, tx in times.items():
        if sec - tx > timeout:
        # if (timedelta(milliseconds=ts) - timedelta(milliseconds=tx)).total_seconds() >= 10:
            todel.append(tinfo)
        else:
            break
    # print(f'Removing {len(todel):,d}', file=stderr)
    for tinfo in todel:
        del times[tinfo]
        del reqs[tinfo]
    return todel

def write_traces(towrite: List[TraceInfo], traces, extra_info: Dict[TraceInfo, Dict[str, int]], outfile, timeout):
    for tinfo in towrite:
        dst = tinfo.dst
        hops = traces.pop(tinfo, None)
        extra = extra_info.pop(tinfo)
        if hops:
            hops = sorted(hops, key=lambda x: x['probe_ttl'])
            hops, stop_reason = truncate_hops(hops, dst)
            if len(hops) > 0:
                hop = hops[0]
                tx = hop['tx']
                if tinfo.proto == IPPROTO_ICMP:
                    method = 'icmp-echo-paris'
                elif tinfo.proto == IPPROTO_UDP:
                    method = 'udp-paris'
                else:
                    continue
                # start = dict(sec=tx['sec'], usec=tx['usec'], ftime='fake')
                start = f'{{"sec": {tx["sec"]}, "usec": {tx["usec"]}, "ftime": "fake"}}'
                truncate_hops(hops, dst)
                output = f'{{"type": "trace", "version": "0.1", "userid": 0, "method": "{method}", "src": "{tinfo.src}", "dst": "{tinfo.dst}", '
                if tinfo.proto == IPPROTO_ICMP:
                    output += f'"icmp_sum": {extra["icmp_sum"]}, '
                elif tinfo.proto == IPPROTO_UDP:
                    output += f'"sport": {extra["sport"]}, "dport": {extra["dport"]}, '
                output += (f'"stop_reason": "{stop_reason}", "stop_data": 0, "start": {start}, "hop_count": {len(hops)}, '
                           f'"attempts": 1, "hoplimit": 32, "firsthop": 1, "wait": {timeout}, "wait_probe": 0, "tos": 0, "probe_size": 64, '
                           f'"hops": {json.dumps(hops)}}}\n')
                outfile.write(output)
                # trace = dict(
                #     type='trace', version='0.1', userid=0,
                #     method=method,
                #     src=tinfo.src, dst=dst,
                #     icmp_sum=47652,
                #     stop_reason=stop_reason, stop_data=0,
                #     start=start, hop_count=len(hops), attempts=1, hoplimit=32, firsthop=1,
                #     wait=10, wait_probe=0, tos=0, probe_size=64, hops=hops
                # )
                # outfile.write(json.dumps(trace) + '\n')

def truncate_hops(hops, dst):
    seen = set()
    prev = None
    for i, hop in enumerate(hops):
        if hop['addr'] == dst:
            return hops[:i+1], 'COMPLETED'
        elif hop['icmp_type'] == 3:
            return hops[:i+1], 'UNREACH'
        addr = hop['addr']
        if addr in seen:
            return hops[:i+1], 'LOOP'
        if prev is not None:
            seen.add(prev)
        prev = addr
    return hops, 'GAPLIMIT'

def read_pcap(infile, outfile, timeout):
    reqs = defaultdict(dict)
    times = OrderedDict()
    traces = defaultdict(list)
    extra_info = {}
    noinfo = 0
    skipped = 0
    wrongtype = 0
    malformed = 0
    start = None
    lines = 0
    pb = Progress(message='Converting pcap to json', increment=100000, callback=lambda: f'Skipped {skipped:,d} NoInfo {noinfo:,d} WrongType {wrongtype:,d} Malformed {malformed:,d}')
    for line in pb.iterator(infile):
        try:
            j = json.loads(line)
            ts = j['timestamp']
            if start is None:
                start = ts
            ip = j['ip']
            proto = ip['ip_proto']
            src = ip['ip_src']
            dst = ip['ip_dst']
            ttl = ip['ip_ttl']
            if proto == IPPROTO_ICMP:
                icmp = j['icmp']
                icmp_type = icmp['icmp_type']
                if icmp_type == 8:
                    tinfo = TraceInfo(src, dst, proto)
                    reqs[tinfo][ttl] = dict(ts_sec=ts['ts_sec'], ts_usec=ts['ts_usec'], src=src, dst=dst, size=ip['ip_len'])
                    extra_info[tinfo] = dict(icmp_sum=icmp['icmp_check'])
                    times[tinfo] = ts['ts_sec']
                else:
                    if icmp_type == 3 or icmp_type == 11:
                        qip = icmp['ip']
                        odst = qip['ip_dst']
                        qproto = qip['ip_proto']
                        if qproto == IPPROTO_ICMP:
                            try:
                                qicmp = icmp['icmp']
                            except KeyError:
                                noinfo += 1
                                continue
                            ottl = qicmp['icmp_seq']
                        elif qproto == IPPROTO_UDP:
                            try:
                                qudp = icmp['udp']
                            except KeyError:
                                noinfo += 1
                                continue
                            ottl = qudp['udp_check']
                        else:
                            wrongtype += 1
                            continue
                        tinfo = TraceInfo(dst, odst, qproto)
                        extra = dict(icmp_q_ttl=qip['ip_ttl'], icmp_q_ipl=qip['ip_len'], icmp_q_tos=qip['ip_tos'])
                    elif icmp_type == 0:
                        ottl = icmp['icmp_seq']
                        odst = src
                        tinfo = TraceInfo(dst, odst, 1)
                        extra = None
                    else:
                        wrongtype += 1
                        continue
                    try:
                        req = reqs[tinfo][ottl]
                    except KeyError:
                        skipped += 1
                        continue
                    tx = timedelta(seconds=req['ts_sec'], microseconds=req['ts_usec'])
                    rx = timedelta(seconds=ts['ts_sec'], microseconds=ts['ts_usec'])
                    rtt = round((rx - tx).total_seconds() * 1000, 3)
                    hop = dict(
                        addr=src, probe_ttl=ottl, probe_id=1, probe_size=req['size'],
                        tx={'sec': req['ts_sec'], 'usec': req['ts_usec']}, rtt=rtt,
                        reply_ttl=ttl, reply_tos=ip['ip_tos'], reply_ipid=ip['ip_id'], reply_size=ip['ip_len'],
                        icmp_type=icmp_type, icmp_code=icmp['icmp_code']
                    )
                    if extra is not None:
                        hop.update(extra)
                    traces[tinfo].append(hop)
            elif proto == IPPROTO_UDP:
                udp = j['udp']
                tinfo = TraceInfo(src, dst, proto)
                reqs[tinfo][ttl] = dict(ts_sec=ts['ts_sec'], ts_usec=ts['ts_usec'], src=src, dst=dst, size=ip['ip_len'])
                extra_info[tinfo] = dict(sport=udp['udp_src'], dport=udp['udp_dst'])
                times[tinfo] = ts['ts_sec']
            lines += 1
            if lines >= 100000:
                todel = remove_old(ts, times, reqs, timeout)
                write_traces(todel, traces, extra_info, outfile, timeout)
                lines = 0
                # removed.update(todel)
        except KeyError:
            # print()
            print(line)
            traceback.print_exc()
            # raise
            malformed += 1
            # raise
        except:
            print()
            print(line)
            raise
    # todel = remove_old(float('inf'), times, reqs)
    write_traces(list(traces.keys()), traces, extra_info, outfile, timeout)

def convert(infile, outfile, timeout, fix=True):
    sidefile = (infile.rpartition('.')[0] if infile.endswith('.gz') or infile.endswith('.bz2') else infile) + '.sidecar'
    if os.path.exists(sidefile):
        with open(sidefile) as f:
            header = f.readline()
            footer = f.readline()
    else:
        header = '{"type":"cycle-start", "list_name":"default", "id":0, "hostname":"fake", "start_time":0}\n'
        footer = '{"type":"cycle-stop", "list_name":"default", "id":0, "hostname":"fake", "stop_time":0}\n'
    if fix:
        pref = f"pcapfix2 -o - {infile}"
    elif infile.endswith('gz'):
        pref = f'gzip -dc {infile}'
    elif infile.endswith('bz2'):
        pref = f'bzip2 -df {infile}'
    else:
        pref = f'cat {infile}'
    cmd = f"{pref} | pcap2json -o - -"
    print(cmd)
    p = Popen(cmd, shell=True, universal_newlines=True, stdout=PIPE)
    fout = fopen(outfile, 'wt') if outfile != '-' else sys.stdout
    try:
        # with fopen(outfile, 'wt') as fout:
        # print(header)
        fout.write(header)
        read_pcap(p.stdout, fout, timeout)
        fout.write(footer)
    finally:
        if outfile == '-':
            fout.close()
    #     p.terminate()
    #     p.communicate()

def convert_wrapper(args):
    infile, outfile = args
    convert(infile, outfile, _timeout, fix=_fix)

def pattern(infile, gzip=False, bzip2=False):
    if infile.endswith('.bz2'):
        outfile = infile[:-4]
    elif infile.endswith('.gz'):
        outfile = infile[:-3]
    else:
        outfile = infile
    outfile = '{}.json'.format(outfile)
    if gzip:
        outfile += '.gz'
    elif bzip2:
        outfile += '.bz2'
    return outfile

def main():
    global _fix, verbose, _timeout
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-r', '--raw', action='store_true')
    parser.add_argument('-v', '--verbose', action='count', default=0)
    parser.add_argument('-o', '--outfiles', nargs='+')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-z', '--gzip', action='store_true')
    group.add_argument('-b', '--bzip2', action='store_true')
    parser.add_argument('-p', '--processes', default=1, type=int)
    parser.add_argument('-i', '--infiles', nargs='+')
    parser.add_argument('-I', '--file')
    parser.add_argument('-t', '--timeout', type=int, default=_timeout)
    args = parser.parse_args()

    _fix = not args.raw
    print('Fixing pcap file: {}'.format(_fix))
    verbose = args.verbose

    _timeout = args.timeout

    infiles = args.infiles if args.infiles is not None else []
    if args.file:
        with open(args.file, 'rt') as f:
            for filename in f:
                infiles.append(filename.strip())

    if args.outfiles:
        if len(args.outfiles) != len(infiles):
            print('Outfiles must match number of infiles.')
            exit(1)
        outfiles = args.outfiles
    else:
        outfiles = []
        for infile in infiles:
            outfile = pattern(infile, gzip=args.gzip, bzip2=args.bzip2)
            outfiles.append(outfile)

    processes = min(args.processes, len(infiles))

    pb = Progress(len(infiles), message=f'Creating traceroutes with {processes} process')
    if processes == 1:
        for infile, outfile in pb.iterator(zip(infiles, outfiles)):
            convert(infile, outfile, _timeout, fix=_fix)
    else:
        Progress.set_output(False)
        with Pool(processes) as pool:
            for _ in pb.iterator(pool.imap_unordered(convert_wrapper, zip(infiles, outfiles))):
                pass
        Progress.set_output(True)

if __name__ == '__main__':
    main()


import os
import subprocess
import tempfile
from argparse import ArgumentParser
from shutil import copyfile

def shuf(input: str, output: str=None, inplace: bool=False, gzip: bool=False, bzip2: bool=False):
    f = None
    fname = None

    try:
        if input.endswith('.gz'):
            infile = 'zcat'
        elif input.endswith('.bz2'):
            infile = 'bzcat'
        else:
            infile = 'cat'
        infile += ' {}'.format(input)

        if inplace:
            f, fname = tempfile.mkstemp()
            outfile = fname
        else:
            outfile = output

        if gzip or outfile.endswith('.gz') or (inplace and input.endswith('.gz')):
            out = '| gzip >'
        elif bzip2 or outfile.endswith('.bz2') or (inplace and input.endswith('.bz2')):
            out = '| bzip2 >'
        else:
            out = '>'
        out = '{} {}'.format(out, outfile)

        # cmd = '{} | shuf {}'.format(infile, out)
        cmd = infile + ' | awk \'BEGIN{srand()}{print rand(), $0}\' | sort -n -k 1 | awk \'sub(/\S* /,"")\' ' + out
        subprocess.run(cmd, shell=True)
        if inplace:
            copyfile(fname, input)
    finally:
        if f is not None and fname is not None:
            os.unlink(fname)

def main():
    parser = ArgumentParser()
    parser.add_argument('-i', '--input')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-I', '--inplace', action='store_true')
    group.add_argument('-o', '--output')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-z', '--gzip', action='store_true')
    group.add_argument('-b', '--bzip2', action='store_true')
    args = parser.parse_args()

    shuf(args.input, output=args.output, inplace=args.inplace, gzip=args.gzip, bzip2=args.bzip2)


if __name__ == '__main__':
    main()



import sys
import time

from setuptools import setup, find_packages
from setuptools.extension import Extension

install_requires = ['scapy', 'file2', 'pb_amarder', 'requests']

if 'build_ext' in sys.argv:
    from Cython.Distutils import build_ext
    use_cython = True
    install_requires += 'cython'
    version = '2.1.'
    version += time.strftime('%Y%m%d%H%M%S', time.gmtime())
    with open('cloudtrace/version.py', 'w') as f:
        f.write(f"__version__ = '{version}'\n")
else:
    use_cython = False
    exec(open('cloudtrace/version.py').read())
    version = __version__

ext_pyx = '.pyx' if use_cython else '.c'
ext_py = '.py' if use_cython else '.c'

extensions_names = {
    'cloudtrace.trace.utils': ['cloudtrace/trace/utils' + ext_pyx],
    'cloudtrace.trace.probe': ['cloudtrace/trace/probe' + ext_pyx],
    'cloudtrace.trace.fasttrace': ['cloudtrace/trace/fasttrace' + ext_py],
    'cloudtrace.trace.scampertrace': ['cloudtrace/trace/scampertrace' + ext_py],
    'cloudtrace.trace.cloudscamper': ['cloudtrace/trace/cloudscamper' + ext_py],
    'cloudtrace.trace.randomize': ['cloudtrace/trace/randomize' + ext_py],

    'cloudtrace.read.utils': ['cloudtrace/read/utils' + ext_pyx],
    'cloudtrace.read.pcap': ['cloudtrace/read/pcap' + ext_pyx],
    'cloudtrace.read.combine': ['cloudtrace/read/combine' + ext_pyx],
    'cloudtrace.read.packet': ['cloudtrace/read/packet' + ext_pyx],
    'cloudtrace.read.convert': ['cloudtrace/read/convert' + ext_pyx],
    'cloudtrace.read.linkedlist': ['cloudtrace/read/linkedlist' + ext_pyx],
    'cloudtrace.read.reader': ['cloudtrace/read/reader' + ext_py],
    'cloudtrace.read.slowread': ['cloudtrace/read/slowread' + ext_py],
}

extensions = [Extension(k, v) for k, v in extensions_names.items()]
package_data = {k: ['*.pxd', '*pyx', '*.py'] for k in extensions_names}

if use_cython:
    from Cython.Build import cythonize
    extensions = cythonize(
        extensions,
        compiler_directives={'language_level': '3', 'embedsignature': True},
        annotate=True,
        gdb_debug=True
    )

# version = '2.1.'
# version += time.strftime('%Y%m%d%H%M%S', time.gmtime())
# with open('cloudtrace/version.py', 'w') as f:
#     f.write(f"__version__ = '{version}'\n")

setup(
    name="cloudtrace",
    # version='REPLACEVERSION',
    version=version,
    author='Alex Marder',
    # author_email='notlisted',
    description="Cloud traceroute.",
    url="https://gitlab.com/alexander_marder/cloudtrace",
    packages=find_packages(),
    # setup_requires=["cython"],
    install_requires=install_requires,
    # cmdclass={'build_ext': build_ext},
    ext_modules=extensions,
    entry_points={
        'console_scripts': [
            'fasttrace=cloudtrace.trace.fasttrace:main',
            'scampertrace=cloudtrace.trace.scampertrace:main',
            'cloudscamper=cloudtrace.trace.cloudscamper:main',
            'cloudshuffle=cloudtrace.trace.shuffle:main',
            'fastread=cloudtrace.read.reader:main',
            'slowread=cloudtrace.read.slowread:main',
            'ct-randomize=cloudtrace.trace.randomize:main'
        ],
    },
    zip_safe=False,
    package_data=package_data,
    include_package_data=True,
    python_requires='>3.6'
)



# import bz2
# import gzip
import json
import os
import platform
import tempfile
import time
import urllib.request
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from datetime import date
from glob import glob
from multiprocessing.context import Process
import socket
from subprocess import Popen, PIPE, run
from time import sleep

# from scapy.config import conf

import cloudtrace.trace.probe
import cloudtrace.read.reader
from cloudtrace.trace.cloudinfo import get_cloud_info
from cloudtrace.trace.shuffle import shuf
from cloudtrace.version import __version__


def new_filename(default_output, proto, pps, ext, gzip=False, bzip2=False):
    # hostname = platform.node()
    hostname = get_cloud_info()
    dirname, basename = os.path.split(default_output)
    if dirname:
        os.makedirs(dirname, exist_ok=True)
    if basename:
        basename += '.'
    timestamp = int(time.time())
    dt = date.fromtimestamp(timestamp)
    datestr = dt.strftime('%Y%m%d')
    filename = os.path.join(dirname, '{base}{host}.{date}.{time}.{proto}.{pps}.{ext}'.format(
        base=basename, host=hostname, date=datestr, time=timestamp, proto=proto, pps=pps, ext=ext))
    if gzip:
        filename += '.gz'
    elif bzip2:
        filename += '.bz2'
    return filename

def remote_notify(pattern, remote):
    # username = getpass.getuser()
    if 'SUDO_USER' in os.environ:
        username = os.environ['SUDO_USER']
    else:
        username = os.environ['USER']
    files = glob(pattern)
    for f in files:
        try:
            host, _, port = remote.partition(':')
            port = int(port)
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((host, port))
            s.send('{}:{}\n'.format(username, f).encode())
            s.close()
        except:
            print('Unable to connect to {}.'.format(remote))
            pass

def craftandsend(targets: str, pid, pps, minttl=1, maxttl=32, proto=1, timer='nano', randomize=False):
    from scapy.config import conf
    iface, src, nexthop = conf.route.route('0.0.0.0')
    # transient.probe.craft(targets, pid, file='test.pcap', minttl=minttl, maxttl=maxttl, proto=proto)
    # cmd = 'sudo tcpreplay -T {} --intf1={} --pps={} -'.format(timer, iface, pps)
    cmd = 'tcpreplay -T {} --intf1={} --pps={} -'.format(timer, iface, pps)
    tcpreplay = Popen(cmd, shell=True, stdin=PIPE, stdout=PIPE)
    cloudtrace.trace.probe.craft(targets, pid, file=tcpreplay.stdin, minttl=minttl, maxttl=maxttl, proto=proto, randomize=randomize)
    tcpreplay.stdin = None
    out, _ = tcpreplay.communicate()
    out = out.decode()
    print(out.strip())

def trace(targets: str, outfile: str, pps, proto, pid, waittime=30, timer='nano', read=False, randomize=False):
    from scapy.config import conf
    iface, src, nexthop = conf.route.route('0.0.0.0')
    comp = None

    if read:
        comp = Popen('fastread -i - -o {}'.format(outfile), shell=True, stdin=PIPE)
    else:
        if outfile.endswith('.gz'):
            comp = Popen('gzip -c > {}'.format(outfile), shell=True, stdin=PIPE)
        elif outfile.endswith('.bz2'):
            comp = Popen('bzip2 -c > {}'.format(outfile), shell=True, stdin=PIPE)
    if comp is None:
        wflag = outfile
        stdout = None
    else:
        wflag = '-'
        stdout = comp.stdin

    if proto == 'icmp':
        send_filter = 'src {} and icmp and icmp[4:2] = {}'.format(src, pid)
    elif proto == 'udp':
        send_filter = 'src {} and udp and udp[0:2] = {}'.format(src, pid)
    else:
        raise NotImplementedError()
    recv_filter = 'icmp and dst {}'.format(src)
    # cmd = r'sudo tcpdump -p -B 40960 -i {iface} -w {wflag} -v \( {send} \) or \( {recv} \)'.format(iface=iface, wflag=wflag, send=send_filter, recv=recv_filter)
    cmd = r'tcpdump -p -B 40960 -i {iface} -w {wflag} -v \( {send} \) or \( {recv} \)'.format(iface=iface, wflag=wflag, send=send_filter, recv=recv_filter)
    print(cmd)
    tcpdump = Popen(cmd, shell=True, stdout=stdout)
    proc = Process(target=craftandsend, args=(targets, pid, pps), kwargs={'proto': proto, 'timer': timer, 'randomize': randomize})
    proc.start()
    proc.join()
    print('Waiting {} seconds'.format(round(waittime)))
    sleep(waittime)
    # run('sudo pkill tcpdump', shell=True)
    run('pkill tcpdump', shell=True)
    tcpdump.wait()
    if comp is not None:
        comp.stdin.flush()
        comp.stdin.close()
        comp.wait()

def main():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-i', '--input')
    group.add_argument('-I', '--addr', nargs='*')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-o', '--output')
    group.add_argument('-d', '--default-output')
    parser.add_argument('-p', '--pps', default=8000, type=int, help='Packets per second.')
    parser.add_argument('-P', '--proto', default='icmp', choices=['icmp', 'udp'], help='Transport protocol.')
    parser.add_argument('--probe-pcap', action='store_true')
    parser.add_argument('--pid', type=int, default=os.getpid())
    parser.add_argument('-w', '--wait', type=float, default=30, help='Seconds to wait after sending all packets')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-z', '--gzip', action='store_true')
    group.add_argument('-b', '--bzip2', action='store_true')
    parser.add_argument('-r', '--remote')
    parser.add_argument('-c', '--cycles', type=int, default=1)
    parser.add_argument('-T', '--timer', default='nano', choices=['nano', 'select', 'ioport', 'gtod'])
    parser.add_argument('-R', '--random', action='store_true')
    parser.add_argument('-S', '--shuffle', action='store_true')
    parser.add_argument('--read', action='store_true')
    # parser.add_argument('--tmp', default='.infile.tmp')
    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=__version__))
    args = parser.parse_args()

    if args.addr:
        f = tempfile.NamedTemporaryFile(mode='wt', delete=False)
        infile = f.name
        for addr in args.addr:
            f.write('{}\n'.format(addr))
        f.close()
    else:
        infile = args.input

    if args.shuffle:
        shuf(infile, inplace=True)

    try:
        cycle = 0
        while args.cycles == 0 or cycle < args.cycles:
            pid = args.pid % 65535
            print('Using pid: {}'.format(pid))

            if args.output:
                filename = args.output
                pattern = filename
            else:
                ext = 'pcap' if not args.read else 'jsonl'
                filename = new_filename(args.default_output, args.proto, args.pps, ext, gzip=args.gzip, bzip2=args.bzip2)
                print('Saving to {}'.format(filename))
                dirname, basename = os.path.split(args.default_output)
                pattern = os.path.join(dirname, '{}.{}*'.format(basename, ext))

            name = platform.node()
            addr = urllib.request.urlopen('https://api.ipify.org/').read().decode('utf8')

            sidefile = (filename.rpartition('.')[0] if filename.endswith('.gz') or filename.endswith('.bz2') else filename) + '.sidecar'
            with open(sidefile, 'wt') as f:
                header = {
                    "type": "cycle-start", "list_name": "default", "id": pid, "hostname": name,
                    "start_time": time.time(), 'addr': addr
                }
                f.write(json.dumps(header) + '\n')

                trace(infile, filename, args.pps, args.proto, pid, waittime=args.wait, timer=args.timer, read=args.read, randomize=args.random)

                stoptime = time.time()

                footer = {
                    "type": "cycle-stop", "list_name": "default", "id": pid, "hostname": name,
                    "stop_time": stoptime, 'addr': addr
                }
                f.write(json.dumps(footer) + '\n')

            if args.remote:
                remote_notify(pattern, args.remote)
            try:
                cycle += 1
            except OverflowError:
                cycle = 1

    finally:
        if args.addr:
            os.unlink(infile)
